@@ -86,7 +86,7 @@ static int open_log(const char *path)
 
 	memset(&addr, 0, sizeof(addr));
 	addr.sun_family = AF_UNIX;
-	strncpy(addr.sun_path, path, sizeof(addr.sun_path));
+	strncpy(addr.sun_path, path, sizeof(addr.sun_path) - 1);
 
 	if (connect(log_fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
 		close_log();
@@ -139,6 +139,7 @@ LIB_EXPORT void l_log_set_null(void)
 	log_func = log_null;
 }
 
+__attribute__((format(printf, 5, 0)))
 static void log_stderr(int priority, const char *file, const char *line,
 			const char *func, const char *format, va_list ap)
 {
@@ -157,6 +158,7 @@ LIB_EXPORT void l_log_set_stderr(void)
 	log_func = log_stderr;
 }
 
+__attribute__((format(printf, 5, 0)))
 static void log_syslog(int priority, const char *file, const char *line,
 			const char *func, const char *format, va_list ap)
 {
@@ -205,6 +207,7 @@ LIB_EXPORT void l_log_set_syslog(void)
 	log_func = log_syslog;
 }
 
+__attribute__((format(printf, 5, 0)))
 static void log_journal(int priority, const char *file, const char *line,
 			const char *func, const char *format, va_list ap)
 {
@@ -369,7 +372,6 @@ LIB_EXPORT void l_debug_add_section(struct l_debug_desc *start,
 					struct l_debug_desc *end)
 {
 	const struct l_queue_entry *entry;
-	const struct debug_section *section;
 	struct debug_section *new_section;
 
 	if (!debug_sections) {
@@ -379,7 +381,7 @@ LIB_EXPORT void l_debug_add_section(struct l_debug_desc *start,
 
 	for (entry = l_queue_get_entries(debug_sections); entry;
 					entry = entry->next) {
-		section = entry->data;
+		const struct debug_section *section = entry->data;
 
 		if (section->start == start && section->end == end)
 			return;
@@ -406,7 +408,6 @@ LIB_EXPORT void l_debug_enable_full(const char *pattern,
 					struct l_debug_desc *end)
 {
 	const struct l_queue_entry *entry;
-	const struct debug_section *section;
 
 	if (!pattern)
 		return;
@@ -417,7 +418,8 @@ LIB_EXPORT void l_debug_enable_full(const char *pattern,
 
 	for (entry = l_queue_get_entries(debug_sections); entry;
 					entry = entry->next) {
-		section = entry->data;
+		const struct debug_section *section = entry->data;
+
 		debug_enable(section->start, section->end);
 	}
 }
@@ -430,11 +432,11 @@ LIB_EXPORT void l_debug_enable_full(const char *pattern,
 LIB_EXPORT void l_debug_disable(void)
 {
 	const struct l_queue_entry *entry;
-	const struct debug_section *section;
 
 	for (entry = l_queue_get_entries(debug_sections); entry;
 					entry = entry->next) {
-		section = entry->data;
+		const struct debug_section *section = entry->data;
+
 		debug_disable(section->start, section->end);
 	}
 
