@@ -24,6 +24,9 @@
 #include <config.h>
 #endif
 
+#include <linux/rtnetlink.h>
+#include <linux/if.h>
+
 #include <ell/ell.h>
 
 #include "linux/nl80211.h"
@@ -49,6 +52,7 @@ struct adhoc_state {
 	struct l_queue *sta_states;
 	uint32_t sta_watch_id;
 	uint32_t netdev_watch_id;
+	unsigned int mlme_watch;
 	struct l_dbus_message *pending;
 	uint32_t ciphers;
 	uint32_t group_cipher;
@@ -131,15 +135,21 @@ static void adhoc_reset(struct adhoc_state *adhoc)
 				dbus_error_aborted(adhoc->pending));
 
 	l_free(adhoc->ssid);
+	adhoc->ssid = NULL;
 
 	netdev_station_watch_remove(adhoc->netdev, adhoc->sta_watch_id);
+	adhoc->sta_watch_id = 0;
 
 	l_queue_destroy(adhoc->sta_states, adhoc_sta_free);
+	adhoc->sta_states = NULL;
 
 	adhoc->started = false;
 
 	l_dbus_property_changed(dbus_get_bus(), netdev_get_path(adhoc->netdev),
 						IWD_ADHOC_INTERFACE, "Started");
+
+	if (adhoc->mlme_watch)
+		l_genl_family_unregister(adhoc->nl80211, adhoc->mlme_watch);
 }
 
 static void adhoc_set_rsn_info(struct adhoc_state *adhoc,
@@ -158,15 +168,31 @@ static bool ap_sta_match_addr(const void *a, const void *b)
 	return !memcmp(sta->addr, b, 6);
 }
 
+static void adhoc_operstate_cb(int error, uint16_t type,
+					const void *data,
+					uint32_t len, void *user_data)
+{
+	if (!error)
+		return;
+
+	l_debug("netdev: %u, error: %s", L_PTR_TO_UINT(user_data),
+							strerror(-error));
+}
+
 static void adhoc_handshake_event(struct handshake_state *hs,
 		enum handshake_event event, void *user_data, ...)
 {
 	struct sta_state *sta = user_data;
 	struct adhoc_state *adhoc = sta->adhoc;
+	va_list args;
+
+	va_start(args, user_data);
 
 	switch (event) {
 	case HANDSHAKE_EVENT_FAILED:
-		l_error("handshake failed with STA "MAC, MAC_STR(sta->addr));
+		l_error("handshake failed with STA "MAC" (%d)",
+						MAC_STR(sta->addr),
+						va_arg(args, int));
 
 		/*
 		 * eapol frees the state machines upon handshake failure. Since
@@ -286,6 +312,7 @@ static void adhoc_start_rsna(struct sta_state *sta, const uint8_t *gtk_rsc)
 	eapol_register(sta->sm_a);
 
 	eapol_start(sta->sm);
+	eapol_start(sta->sm_a);
 
 	return;
 
@@ -395,7 +422,15 @@ static void adhoc_new_station(struct adhoc_state *adhoc, const uint8_t *mac)
 
 	/* with open networks nothing else is required */
 	if (sta->adhoc->open) {
+		int ifindex = netdev_get_ifindex(adhoc->netdev);
+
 		sta->authenticated = true;
+
+		l_rtnl_set_linkmode_and_operstate(iwd_get_rtnl(), ifindex,
+					IF_LINK_MODE_DORMANT, IF_OPER_UP,
+					adhoc_operstate_cb,
+					L_UINT_TO_PTR(ifindex), NULL);
+
 		l_dbus_property_changed(dbus_get_bus(),
 					netdev_get_path(adhoc->netdev),
 					IWD_ADHOC_INTERFACE, "ConnectedPeers");
@@ -463,11 +498,35 @@ static void adhoc_join_cb(struct netdev *netdev, int result, void *user_data)
 
 	reply = l_dbus_message_new_method_return(adhoc->pending);
 	dbus_pending_reply(&adhoc->pending, reply);
+}
 
-	adhoc->started = true;
+static void adhoc_mlme_notify(struct l_genl_msg *msg, void *user_data)
+{
+	struct adhoc_state *adhoc = user_data;
+	uint32_t ifindex;
 
-	l_dbus_property_changed(dbus_get_bus(), netdev_get_path(adhoc->netdev),
+	if (nl80211_parse_attrs(msg, NL80211_ATTR_IFINDEX, &ifindex,
+				NL80211_ATTR_UNSPEC) < 0 ||
+			ifindex != netdev_get_ifindex(adhoc->netdev))
+		return;
+
+	switch (l_genl_msg_get_command(msg)) {
+	case NL80211_CMD_JOIN_IBSS:
+		/*
+		 * if watch is set the join_ibss_cb has come back. This event
+		 * will come in for each new STA joining the IBSS so we only
+		 * want to set it once
+		 */
+		if (adhoc->sta_watch_id && !adhoc->started) {
+			adhoc->started = true;
+
+			l_dbus_property_changed(dbus_get_bus(),
+						netdev_get_path(adhoc->netdev),
 						IWD_ADHOC_INTERFACE, "Started");
+		}
+
+		break;
+	}
 }
 
 static struct l_dbus_message *adhoc_dbus_start(struct l_dbus *dbus,
@@ -508,6 +567,11 @@ static struct l_dbus_message *adhoc_dbus_start(struct l_dbus *dbus,
 			adhoc))
 		return dbus_error_invalid_args(message);
 
+	adhoc->mlme_watch = l_genl_family_register(adhoc->nl80211, "mlme",
+						adhoc_mlme_notify, adhoc, NULL);
+	if (!adhoc->mlme_watch)
+		return dbus_error_failed(message);
+
 	return NULL;
 }
 
@@ -543,6 +607,11 @@ static struct l_dbus_message *adhoc_dbus_start_open(struct l_dbus *dbus,
 			adhoc))
 		return dbus_error_invalid_args(message);
 
+	adhoc->mlme_watch = l_genl_family_register(adhoc->nl80211, "mlme",
+						adhoc_mlme_notify, adhoc, NULL);
+	if (!adhoc->mlme_watch)
+		return dbus_error_failed(message);
+
 	return NULL;
 }
 
@@ -576,9 +645,11 @@ static struct l_dbus_message *adhoc_dbus_stop(struct l_dbus *dbus,
 	if (!adhoc->started)
 		return l_dbus_message_new_method_return(message);
 
-	if (!netdev_leave_adhoc(adhoc->netdev, adhoc_leave_cb, adhoc))
+	if (netdev_leave_adhoc(adhoc->netdev, adhoc_leave_cb, adhoc))
 		return dbus_error_failed(message);
 
+	adhoc->pending = l_dbus_message_ref(message);
+
 	return NULL;
 }
 
@@ -700,3 +771,4 @@ static void adhoc_exit(void)
 }
 
 IWD_MODULE(adhoc, adhoc_init, adhoc_exit)
+IWD_MODULE_DEPENDS(adhoc, netdev);
